#!/usr/local/bin/ruby -Ks

# Copyright (C) 2005 Okada Jun All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are 
# met: 
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer 
#    as the first lines of this file unmodified.
#
# 2. Redistributions in  binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#   documentation and/or other materials provided with the distribution. 
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
# DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, 
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.


$KCODE = 'sjis'
DEFINE_FILENAME = 'updater.define.txt'
CONFIG_FILENAME = 'updater.conf'
FILELIST_FILENAME = 'updater.filelist.txt'
TMP_PATH = 'updatertmp/'


require 'vr/vruby'
require 'vr/vrcontrol'
require 'vr/vrcomctl'
require 'vr/vrdialog'
require 'net/http'
require 'digest/md5'
require 'uri'
require 'fileutils'

#require 'exerb/exerb'
require 'exerb/version'
require 'exerb/archive'
require 'exerb/executable'
require 'exerb/recipe'
require 'exerb/resource'
require 'exerb/resource_library'

require 'win32'
require 'messagebox'
require 'kconv'


# XNvg݂̑fBNgɈړ
SELF_PATH = (defined?(Exerb) && Exerb.methods.include?("runtime?") ? Exerb.filepath : $0)
BASE_DIR  = File.dirname(SELF_PATH.gsub(/\\/, '/'))
Dir.chdir(BASE_DIR)


include MessageBox
Net::HTTP.version_1_2 


module WStyle
  BS_DEFPUSHBUTTON = 1
end

# _CWFXgɁA傫t@Cǂ݂Ȃ
# _CWFXgvZ֐ǉ
class Digest::Base
  def self.open(path)
    obj = new
    
    File.open(path, 'rb') {|f|
      buf = ""
      while f.read(1024 * 1024, buf)
	obj << buf
      end
    }
    obj
  end
end

# httpIuWFNg擾
def http
  if $config[:use_proxy]
    Net::HTTP::Proxy($config[:proxy_host], $config[:proxy_port].to_i == 0 ? 80 : $config[:proxy_port].to_i)
  else
    Net::HTTP
  end
end

# t@C\
FileInfo = Struct.new("FileInfo", :srcFile, :dstFile, :descFile, :size, :md5sum)

class FileInfo
  def to_a
    [ srcFile, dstFile, descFile, size, md5sum ]
  end

  def tmpFilename
    TMP_PATH + dstFile
  end
end


class Array
  def to_file
    info = FileInfo.new
    info.srcFile, info.dstFile, info.descFile, info.size, info.md5sum = *self
    info.size = info.size.to_i
    info
  end
end


# XN[o[teLXg{bNX  
class DescriptionForm < VRModelessDialog
  def construct
    # vpeB̏
    self.caption     = $define.caption + ': XVe'

    # TCYݒ
    w, h = 400, 300
    self.move(@screen.w / 2 - w / 2, @screen.h / 2 - h / 2, w, h)

    addControl(VRText, "text", "", 10, 10, w - 20 - 4, 200, WStyle::WS_VSCROLL)
    addControl(VRButton, "ok", "OK", (w - 70) / 2 - 50, 230, 70, 20, WStyle::WS_TABSTOP | WStyle::BS_DEFPUSHBUTTON)
    addControl(VRButton, "cancel", "LZ", (w - 70) / 2 + 50, 230, 70, 20, WStyle::WS_TABSTOP)
    
    font = @screen.factory.newfont("lrSVbN", 12)
    @text.setFont(font)
    font = @screen.factory.newfont("Terminal", 14)
    @ok.setFont(font)
    @cancel.setFont(font)
    @text.readonly=true
    @ok.focus
    @ok_pressed = false
  end

  def self_wmcommand(msg)
    case LOWORD(msg.wParam)
    when ID_OK, ID_CANCEL
      close(0)
    else
      super(msg)
    end
  end

  def ok?
    @ok_pressed
  end

  def ok_clicked
    @ok_pressed = true
    close(0)
  end

  def cancel_clicked
    close(0)
  end

  def describe(filelist)
    texts = []
    filelist.each do |info|
      text = "#{info.dstFile}: #{info.size}bytes\r\n"
      unless info.descFile.empty?
	text += Kconv.tosjis(get_remote(info.descFile))
      end
      text.strip!
      texts << text
    end
    @text.text = texts.join("\r\n\r\n")

    VRLocalScreen.messageloop
  end
end


# _E[hpl
class DownloadProgressPanel < VRPanel
  include VRStdControlContainer

  def construct
    addControl(VRStatic, 'label', '', 0, 0, 280, 14)
    addControl(VRProgressbar, 'progress', '', 0, 14, 280, 20)
    font = @screen.factory.newfont("Terminal", 14)
    @label.setFont(font)
    @progress.setRange(0, 100)
    @prev_time = 0
  end

  # isxݒ肷
  def setProgress(name, curSize, fullSize, force = false)
    now = Time.now.to_f
    return if now - @prev_time < 0.2 && ! force
    @prev_time = now
    @progress.position = curSize * 100 / fullSize
    @label.caption = "#{name}: #{curSize}/#{fullSize}Kbytes"
  end
end


# _E[hpl 
class DownloadForm < VRModelessDialog
  def construct
    # vpeB̏
    self.caption     = $define.caption + ': XV'

    # TCYݒ
    w, h = 300, 160
    self.move(@screen.w / 2 - w / 2, @screen.h / 2 - h / 2, w, h)

    addControl(DownloadProgressPanel, "afile", "", 5, 5, 280, 34)
    addControl(DownloadProgressPanel, "allfile", "", 5, 49, 280, 34)
    addControl(VRButton, "cancel", "LZ" , (300 - 70) / 2, 100, 70, 20)

    # setting font to controls.
    font = @screen.factory.newfont("Terminal", 14)
    @cancel.setFont(font)
  end

  def cancel_clicked
    @dlthread.kill
    @canceled = true
    close(0)
  end

  # _E[hs
  def download(filelist)
    @dlthread = Thread.start do
      totalSize = filelist.inject(0){|a, info| a + info.size}
      totalSizeDown = 0
      filelist.each do |info|
        FileUtils.mkdir_p(File.dirname(info.tmpFilename))
        uri = URI.parse(remote_path(info.srcFile))
        3.times do
          begin
            resp = nil
            File.open(info.tmpFilename, 'wb') do |file|
              http.start(uri.host, uri.port) do |htt|
                fileSize = info.size
                fileSizeDown = 0
                resp = htt.get(uri.path) do |block|
                  file.write(block)
                  fileSizeDown += block.size
                  totalSizeDown += block.size
                  @afile.setProgress(info.srcFile, fileSizeDown, fileSize)
                  @allfile.setProgress('total', totalSizeDown, totalSize)
                end
                @afile.setProgress(info.srcFile, fileSizeDown, fileSize, true)
              end
            end

            if %w(301 302 303).include?(resp.code)
              if resp['location'] =~ %r!^https?://!
                uri = URI.parse(resp['location'])
              else
                uri.path = resp['location']
              end
              totalSizeDown -= fileSizeDown
            else
              break
            end

          rescue
            next
          end
          @allfile.setProgress('total', totalSizeDown, totalSize, true)
        end
      end
      close(0)
    end
    VRLocalScreen.messageloop
    raise if @canceled
  end
end

# ݒpl
class ConfigForm < VRModelessDialog
  include VRStdControlContainer

  def addAutoUpdate(x, y, w, h)
    @auto_update = true

    addControl(VRGroupbox, 'auto_update_box', 'XV̐ݒ', x, y, w, h, WStyle::WS_GROUP)
    addControl(VRRadiobutton, 'auto_update_off', 'XVȂ', x + 10, y + 20, 150, 14)
    addControl(VRRadiobutton, 'auto_update_on', 'XV', x + 10, y + 40, 150, 14, 0)

    font = @screen.factory.newfont("Terminal", 14)
    @auto_update_box.setFont(font)
    @auto_update_on.setFont(font)
    @auto_update_off.setFont(font)
  end

  def addProxy(x, y, w, h)
    @proxy = true

    @proxy_ctrls = []

    addControl(VRGroupbox, 'proxy_box', 'vNV̐ݒ', x, y, w, h, WStyle::WS_GROUP)
    addControl(VRRadiobutton, 'proxy_off', 'ڃT[oɐڑ', x + 10, y + 20, 150, 14)
    addControl(VRRadiobutton, 'proxy_on', 'vNVoRŃT[oɐڑ', x + 10, y + 40, 200, 14)
    @proxy_ctrls << addControl(VRStatic, 'proxy_label_host', 'zXg', x + 10, y + 60, 150, 14)
    @proxy_ctrls << addControl(VREdit, 'proxy_host', '', x + 10, y + 80, 250, 16, WStyle::WS_TABSTOP)
    @proxy_ctrls << addControl(VRStatic, 'proxy_label_port', '|[g', x + 10, y + 100, 150, 14)
    @proxy_ctrls << addControl(VREdit, 'proxy_port', '', x + 10, y + 120, 60, 16, WStyle::WS_TABSTOP)

    font = @screen.factory.newfont("Terminal", 14)
    @proxy_box.setFont(font)
    @proxy_on.setFont(font)
    @proxy_off.setFont(font)
    @proxy_ctrls.each {|item| item.setFont(font) }
  end

  def proxy_off_clicked
    @proxy_ctrls.each do |item|
      item.enabled = false
    end
  end

  def proxy_on_clicked
    @proxy_ctrls.each do |item|
      item.enabled = true
    end
  end

  def self_wmcommand(msg)
    case LOWORD(msg.wParam)
    when ID_OK
      ok_clicked
    when ID_CANCEL
      cancel_clicked
    else
      super(msg)
    end
  end

  def config
    if @auto_update
      # XVݒ
      if $config[:auto_update]
	@auto_update_on.check(true)
      else
	@auto_update_off.check(true)
      end
    end

    if @proxy
      # vNVݒ
      if $config[:use_proxy]
	@proxy_on.check(true)
	proxy_on_clicked
      else
	@proxy_off.check(true)
      proxy_off_clicked
      end

      @proxy_host.text = $config[:proxy_host]
      @proxy_port.text = $config[:proxy_port]
    end

    VRLocalScreen.messageloop
  end

  def write_back
    if @auto_update
      $config[:auto_update] = @auto_update_on.checked?
    end
    if @proxy
      $config[:use_proxy] = @proxy_on.checked?
      $config[:proxy_host] = @proxy_host.text
      $config[:proxy_port] = @proxy_port.text
    end
  end

  def ok_clicked
    write_back
    write_config
    close(0)
  end
  
  def cancel_clicked
    close(0)
  end

  def update_clicked
    write_back
    $update_now = true
    close(0)
  end
end
    
class ConfigForm0 < ConfigForm
  def construct
    # vpeB̏
    self.caption     = $define.caption + ': lbg[Nݒ'

    # TCYݒ
    w, h = 300, 240
    self.move(@screen.w / 2 - w / 2, @screen.h / 2 - h / 2, w, h)

    # Rg[̒ǉ
    addProxy(10, 10, 275, 150)

    addControl(VRButton, "ok", "OK" , (w / 2 - 70 / 2), 180, 70, 20, WStyle::BS_DEFPUSHBUTTON | WStyle::WS_TABSTOP)

    font = @screen.factory.newfont("Terminal", 14)
    @ok.setFont(font)
    @ok.focus
  end
end

class ConfigForm1 < ConfigForm
  def construct
    # vpeB̏
    self.caption     = $define.caption + ': lbg[Nݒ'

    # TCYݒ
    w, h = 300, 320
    self.move(@screen.w / 2 - w / 2, @screen.h / 2 - h / 2, w, h)

    # Rg[̒ǉ
    addAutoUpdate(10, 10, 275, 65)
    addProxy(10, 90, 275, 150)

    addControl(VRButton, "update", "ɍXV" , (w / 2 - 70 / 2) - 90, 260, 70, 20, WStyle::WS_TABSTOP)
    addControl(VRButton, "ok", "OK" , (w / 2 - 70 / 2), 260, 70, 20, WStyle::BS_DEFPUSHBUTTON | WStyle::WS_TABSTOP)
    addControl(VRButton, "cancel", "LZ" , (w / 2 - 70 / 2) + 90, 260, 70, 20, WStyle::WS_TABSTOP)

    font = @screen.factory.newfont("Terminal", 14)
    @update.setFont(font)
    @ok.setFont(font)
    @cancel.setFont(font)
    @ok.focus
  end
end
  

# RtBOt@Cǂݍ
def read_config
  begin 
    $config = Marshal.load(File.read(CONFIG_FILENAME))
    $first_launch = false
  rescue
    $config = {}
    $first_launch = true
  end

  $config[:auto_update] = true unless $config.include?(:auto_update)
  $config[:use_proxy] = false unless $config.include?(:use_proxy)
  $config[:proxy_host] = '' unless $config.include?(:proxy_host)
  $config[:proxy_port] = '' unless $config.include?(:proxy_port)
end

# RtBOt@Co
def write_config
  File.open(CONFIG_FILENAME, 'wb') do |file|
    Marshal.dump($config, file)
  end
end
   

# `t@C̗L𔻒肷
def define_exist?
  File.file?(DEFINE_FILENAME)
end

# `t@Cǂݍ
def read_define
  return $define if $define
  $define = {}
  src = open(DEFINE_FILENAME) { |f| f.read }
  src.gsub!(/^#.*/, '')
  src.scan(/(\w+?):\s(.+?)(?=^\w*?:\s|\Z)/m) do |key, value|
    $define[key.strip] = value.strip
  end

  raise "specify url." unless $define['url']
  raise "specify caption."   unless $define['caption']
  raise "specify first_lauch." unless $define['first_launch']

  def $define.method_missing(name, *args)
    return self[name.to_s] if self.include?(name.to_s)
    super.method_missing(name, *args)
  end

  $urls = $define['url'].split
end

# T[oURL𓾂
def remote_path(path)
  $url_count ||= rand($urls.size)
  result = $urls[$url_count] + '/' + path
  $url_count = ($url_count + 1) % $urls.size
  result
end

# T[ot@C擾
def get_remote(path)
  uri = URI.parse(remote_path(path))
  3.times do 
    begin
      resp = nil
      http.start(uri.host, uri.port) do |htt|
        resp = htt.get(uri.path)
      end

      if %w(301 302 303).include?(resp.code)
        if resp['location'] =~ %r!^https?://!
          uri = URI.parse(resp['location'])
        else
          uri.path = resp['location']
        end
      else
        return resp.body
      end
    rescue
      next
    end
  end
  raise
end

# t@CXg͂FileInfo̔zԂ
def parse_filelist(txt)
  datum = []
  txt.each do |line|
    items = line.split(',').collect{|s| s.strip}
    next if items.size != 5
    datum << items.to_file
  end
  datum
end

# T[õt@CXg擾
def get_remote_filelist
  parse_filelist(get_remote(FILELIST_FILENAME))
end

# [J̃t@CXg擾
def get_local_filelist
  return [] unless File.file?(FILELIST_FILENAME)
  parse_filelist(File.read(FILELIST_FILENAME))
end

# [J̃t@CXgo
def put_local_filelist(filelist)
  File.open(FILELIST_FILENAME, 'w') do |file|
    filelist.each do |info|
      file.puts info.to_a.join(',')
    end
  end
end

# t@CXXg̐؂
def validate_filelist(filelist)
  pwd = Dir.pwd
  filelist.each do |info|
    path = File.expand_path(info.dstFile)
    # updater.exȇ݂tH_̉ʃtH_ւ̃t@C̓]Ȃ
    unless path[0, pwd.size] == pwd
      messageBox(<<-"END",
T[õt@C₂Ă\܂I
ɃT[o̊Ǘ҂ɘAĉB
		 END
		 'xI')
      exit
    end
  end
end

# ꎞt@C̏ڍׂ
def describe_files(filelist)
  form = VRLocalScreen.modelessform(nil, nil, DescriptionForm)
  form.describe(filelist)
  form.ok?
end

# ꎞt@CT[o_E[h
def fetch_tmpfiles(filelist)
  return if filelist.empty?
  form = VRLocalScreen.modelessform(nil, nil, DownloadForm)
  form.download(filelist)
end

# _E[hłȂꎞt@CԂ
def invalid_tmpfiles(filelist)
  files = []
  filelist.each do |info|
    if ! File.exist?(info.tmpFilename) ||
	info.md5sum != Digest::MD5.open(info.tmpFilename).to_s
      files << info
    end
  end
  files
end

# ꎞt@C𐳋K̃t@Cɒu
def rename_tmpfiles(filelist)
  filelist.each do |info|
    FileUtils.mkdir_p(File.dirname(info.dstFile))
    FileUtils.mv(info.tmpFilename, info.dstFile)
  end
end

# ꎞfBNg폜
def delete_tmpdir
  FileUtils.rm_r(TMP_PATH) if File.exist?(TMP_PATH)
end

# t@CXg}[W
def merge_filelist(srcList, dstList)
  result = dstList.dup
  result.reject! do |dstInfo|
    srcList.find{|srcInfo| srcInfo.srcFile == dstInfo.srcFile && srcInfo.dstFile == dstInfo.dstFile }
  end
  result + srcList
end

# sǂK˂
def query_auto_update
  messageBox($define.first_launch,
	     $define.caption + ': XV',
	     MB_YESNO | MB_ICONQUESTION) == ID_YES
end

# XVsǂ肷
def auto_update?
  if $first_launch
    $config[:auto_update] = query_auto_update
    write_config
    edit_config_first if $config[:auto_update]
  end

  $config[:auto_update]
end

# t@CXV
def update_files(force_dialog = false)
  begin
    # [gA[J̃t@CXg擾
    remote_filelist = get_remote_filelist
    local_filelist = get_local_filelist

    # T[õt@CXg̐`FbN
    validate_filelist(remote_filelist)
    
    # tFb`t@C̃Xg쐬
    fetch_filelist = []
    remote_filelist.each do |info|
      unless local_filelist.include?(info)
      fetch_filelist << info
      end
    end
    
    if fetch_filelist.empty?
      if force_dialog
	messageBox("Abvf[g܂", $define.caption + ': Abvf[g')
      end
      return 
    end

    # ڍׂmF
    raise unless describe_files(fetch_filelist)
    
    # ꎞt@C_E[hāAl[
    not_fetched_filelist = fetch_filelist
    loop do
      fetch_tmpfiles(not_fetched_filelist)
      not_fetched_filelist = invalid_tmpfiles(not_fetched_filelist)
      break if not_fetched_filelist.empty?
      if messageBox(not_fetched_filelist.collect{|info| info.dstFile}.join("\n") + 
		    "\ñ_E[hɎs܂\nx_E[h݂܂H",
		    $define.caption + ': _E[hs',
		    MB_YESNO | MB_ICONEXCLAMATION) == ID_NO
	raise
      end
    end
    
    rename_tmpfiles(fetch_filelist)
    local_filelist = merge_filelist(fetch_filelist, local_filelist)
    
    # [J̃t@CXgXV
    put_local_filelist(local_filelist)
    
    # XV
    messageBox("Abvf[g܂", $define.caption + ': Abvf[g')
  rescue
    messageBox("lbg[NAbvf[gɎs܂",  $define.caption + ': Abvf[gs', MB_ICONEXCLAMATION)
  ensure
    delete_tmpdir
  end
end

# vOs
def start_program
  if $define.include?('exec')
    exec($define.exec)
  end
end

# ݒύX
def edit_config_first
  form = VRLocalScreen.modelessform(nil, nil, ConfigForm0)
  form.config
end

# ݒύX
def edit_config
  form = VRLocalScreen.modelessform(nil, nil, ConfigForm1)
  form.config
end

# RtBOݒ肷
def config
  read_config
  read_define
  begin 
    $update_now = false
    edit_config
    update_files(true) if $update_now
  end while $update_now 
end  

# t@CXV
def update
  read_config
  read_define
  update_files if auto_update?
  start_program
end

# t@CXg\z
def list(filenames)
  new_filelist = []
  new_filenames = []
  base_path = File.expand_path(BASE_DIR)
  filenames.each do |filename|
    file_path = File.expand_path(filename)
    next unless base_path == file_path[0, base_path.size]
    filename = file_path[(base_path.size + 1)..-1]
    next unless File.file?(filename)
    new_filenames << filename
    info = FileInfo.new
    info.srcFile = filename
    info.dstFile = filename
    descFile = File.basename(filename, '.*') + '.txt'
    info.descFile = File.file?(descFile) && descFile != filename ? descFile : ''
    info.size = File.stat(filename).size
    info.md5sum = Digest::MD5.open(filename).to_s
    new_filelist << info
  end
  filelist = get_local_filelist
  filelist = merge_filelist(new_filelist, filelist)
  put_local_filelist(filelist)
  messageBox(new_filenames.join("\r\n") + "\r\nt@CXgɉ܂", 
	     'XgXV')
end

# wv\
def help
  exe_name = File.basename(SELF_PATH)
  messageBox(<<-"END",
#{exe_name} new_files..
    * Abv[htH_pw
    Vt@Ct@CXgɒǉ܂B
    t@CR}hCŎw肷ւɁA
    hbOhbvĂnjłB
    łɃt@CXgɑ݂t@Cw肵ɂ́A
    ̃t@C̃`FbNTXV܂B

#{exe_name} -setup
    * _E[htH_pw
    lbg[Nݒs_CAO\܂B
    K؂ɐݒ肳ꂽ updater.define.txt KvłB

#{exe_name}
    * _E[htH_pw
    lbg[NAbvf[gs܂B
    K؂ɐݒ肳ꂽ updater.define.txt KvłB

#{exe_name} -help
    ̃wv\܂
	     END
	     'wv')
end

# C[`
def main
  if ARGV[0] == '-setup'
    if define_exist?
      config
    else
      help
    end
  elsif ARGV[0] == '-help'
    help
  elsif ! ARGV.empty?
    list(ARGV)
  else
    if define_exist?
      update
    else
      help
    end
  end
end

main


